#include"bbb.h"

#include<stddef.h>
#include"host.h"
#include"../sys.h"
#include"../std.h"

#define SCSI_OP_INQUIRY 0x12
#define SCSI_OP_TEST_UNIT_READY 0x00
#define SCSI_OP_REQUEST_SENSE 0x03
#define SCSI_OP_READ_CAPACITY_10 0x25
#define SCSI_OP_READ_10 0x28
#define SCSI_OP_MODE_SENSE_6 0x1A

typedef struct __attribute__((packed)) {
	uint8_t op;
	union __attribute__((packed)) {
		struct __attribute__((packed)) {
			uint8_t lun5;
			uint8_t page;
			uint8_t reserved;
			uint8_t dataSz;
			uint8_t control;
		} inquiry;
		struct __attribute__((packed)) {
			uint8_t res0;
			uint8_t res1;
			uint8_t res2;
			uint8_t res3;
			uint8_t control;
		} testUnitReady;
		struct __attribute__((packed)) {
			uint8_t desc;
			uint8_t res0;
			uint8_t res1;
			uint8_t dataSz;
			uint8_t control;
		} requestSense;
		struct __attribute__((packed)) {
			uint8_t res0;
			uint8_t res1;
			uint8_t res2;
			uint8_t res3;
			uint8_t res4;
			uint8_t res5;
			uint8_t res6;
			uint8_t res7;
			uint8_t control;
		} readCapacity10;
		struct __attribute__((packed)) {
			uint8_t flags;
			uint32_t lba;
			uint8_t group;
			uint16_t lbaCount;
			uint8_t control;
		} read10;
		struct __attribute__((packed)) {
			uint8_t flags;
			uint8_t pageCode;
			uint8_t subpageCode;
			uint8_t allocationSz;
			uint8_t control;
		} modeSense6;
	};
} SCSICommand;

#define CBW_SIGN 0x43425355
#define CBW_FLAG_IN 0x80
#define CBW_FLAG_OUT 0x00
typedef struct __attribute__((packed)) {
	uint32_t signature;
	uint32_t tag;
	uint32_t dataSz;
	uint8_t flags;
	uint8_t lun;
	uint8_t cmdSz;
	union {
		SCSICommand cmd;
		uint8_t padding[16];
	};
} CBW;

#define CSW_SIGN 0x53425355
typedef struct __attribute__((packed)) {
	uint32_t signature;
	uint32_t tag;
	uint32_t residue;
	uint8_t status;
} CSW;

void dbg(int j, uint8_t b);
extern uint32_t asdfasdf;

static int BBBRequest(USBDevice *dev, CBW cbw, void *buf, size_t len, CSW *ret) {
	if(HostSendBulk(dev, dev->bbb.maxPacketSize, &dev->bbb.toggleOut, dev->bbb.endpOut, &cbw, sizeof(cbw)) == 0) {
		if(BBBReset(dev) == 0) {
			return 0;
		}
		if(HostSendBulk(dev, dev->bbb.maxPacketSize, &dev->bbb.toggleOut, dev->bbb.endpOut, &cbw, sizeof(cbw)) == 0) {
			return 0;
		}
	}
	
	sys_sleep(10);

	if(len) {
		HostSendBulk(dev, dev->bbb.maxPacketSize,
			(cbw.flags & CBW_FLAG_IN) ? &dev->bbb.toggleIn : &dev->bbb.toggleOut,
			(cbw.flags & CBW_FLAG_IN) ? dev->bbb.endpIn : dev->bbb.endpOut,
			buf, len);
	}
	
	return HostSendBulk(dev, dev->bbb.maxPacketSize, &dev->bbb.toggleIn, dev->bbb.endpIn, ret, sizeof(*ret));
}

int BBBReset(USBDevice *dev) {
	if(!HostSendRequest(dev, NULL, (USBStandardRequest) {
		.requestType = 0x21,
		.request = 0xFF, // mass storage reset
		.value = 0,
		.index = 0,
		.length = 0
	})) return 0;
	sys_sleep(150);
	if(!HostSendRequest(dev, NULL, (USBStandardRequest) {
		.requestType = 2, // endpoint
		.request = 1, // clear feature
		.value = 0, // halt
		.index = dev->bbb.endpIn,
		.length = 0
	})) return 0;
	sys_sleep(150);
	if(!HostSendRequest(dev, NULL, (USBStandardRequest) {
		.requestType = 2, // endpoint
		.request = 1, // clear feature
		.value = 0, // halt
		.index = dev->bbb.endpOut,
		.length = 0
	})) return 0;
	sys_sleep(150);
	dev->bbb.toggleIn = dev->bbb.toggleOut = 0;
	return 1;
}

int BBBInquiry(USBDevice *dev) {
	uint8_t inquiry[36];
	
	CSW csw;
	
	return BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = 36,
		.flags = CBW_FLAG_IN,
		.lun = 0,
		.cmdSz = 6,
		.cmd = {
			.op = SCSI_OP_INQUIRY,
			.inquiry = {
				.lun5 = 0,
				.page = 0,
				.reserved = 0,
				.dataSz = 36,
				.control = 0
			}
		}
	}, inquiry, 36, &csw) && csw.status == 0;
}

int BBBTestUnitReady(USBDevice *dev) {
	CSW csw;
	
	return BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = 0,
		.flags = CBW_FLAG_OUT,
		.lun = 0,
		.cmdSz = 6,
		.cmd = {
			.op = SCSI_OP_TEST_UNIT_READY,
			.testUnitReady = {}
		}
	}, NULL, 0, &csw) && csw.status == 0;
}

int BBBRequestSense(USBDevice *dev) {
	uint8_t data[18];
	
	CSW csw;
	
	return BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = 18,
		.flags = CBW_FLAG_IN,
		.lun = 0,
		.cmdSz = 6,
		.cmd = {
			.op = SCSI_OP_REQUEST_SENSE,
			.requestSense = {.dataSz = 18}
		}
	}, data, 18, &csw) && csw.status == 0;
}

int BBBReadCapacity10(USBDevice *dev) {
	uint32_t data[2];
	CSW csw;
	if(!BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = sizeof(data),
		.flags = CBW_FLAG_IN,
		.lun = 0,
		.cmdSz = 10,
		.cmd = {
			.op = SCSI_OP_READ_CAPACITY_10,
			.readCapacity10 = {}
		}
	}, data, sizeof(data), &csw)) return 0;
	
	if(csw.status) return 0;
	
	dev->bbb.sectorSize = BIG_TO_HOST_ENDIAN32(data[1]);
	
	//dbg(asdfasdf + 2, csw.status);
	//dbg(asdfasdf + 3, dev->bbb.maxPacketSize);
	//dbg(asdfasdf + 4, dev->bbb.sectorSize >> 9);
	
	return 1;
}

int BBBRead10(USBDevice *dev, uint32_t lbaStart, void *buf, uint16_t lbaCount) {
	CSW csw;
	if(!BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = lbaCount * dev->bbb.sectorSize,
		.flags = CBW_FLAG_IN,
		.lun = 0,
		.cmdSz = 10,
		.cmd = {
			.op = SCSI_OP_READ_10,
			.read10 = {.lba = BIG_TO_HOST_ENDIAN32(lbaStart), .lbaCount = BIG_TO_HOST_ENDIAN16(lbaCount)}
		}
	}, buf, lbaCount * dev->bbb.sectorSize, &csw)) return 0;
	
	return csw.status == 0;
}

int BBBDeviceResponse(USBDevice *dev, void *area, uint32_t meta) {
	uint64_t start = ((uint64_t*) area)[0];
	uint32_t length = ((uint32_t*) area)[2];
	
	uint32_t startLBA = start / dev->bbb.sectorSize;
	uint32_t endLBA = (start + length - 1) / dev->bbb.sectorSize;
	
	dbg(asdfasdf + 10, startLBA & 0xFF);
	dbg(asdfasdf + 11, endLBA & 0xFF);
	
	if(!BBBRead10(dev, startLBA, area, endLBA - startLBA + 1)) {
		dbg(asdfasdf + 12, 0xFF);
		sys_signal_send(area, 0);
	} else {
		dbg(asdfasdf + 12, 0x00);
		std_move(&area[0], &area[start % dev->bbb.sectorSize], length);
		sys_signal_send(area, 1);
	}
	
	return 1;
}

int BBBModeSense(USBDevice *dev) {
	uint8_t data[192];
	
	CSW csw;
	
	return BBBRequest(dev, (CBW) {
		.signature = CBW_SIGN,
		.tag = 0,
		.dataSz = 192,
		.flags = CBW_FLAG_IN,
		.lun = 0,
		.cmdSz = 6,
		.cmd = {
			.op = SCSI_OP_MODE_SENSE_6,
			.modeSense6 = {
				.pageCode = 0x3F,
				.allocationSz = 192
			}
		}
	}, data, 192, &csw) && csw.status == 0;
}
