#include"usb.h"

#include"host.h"
#include"bbb.h"
#include"../sys.h"

void dbg(int j, uint8_t b);
extern uint32_t asdfasdf;

int USBDeviceDiscover(USBDevice *dev) {
	uint8_t tmp = dev->id;
	dev->id = 0; // 0 is the correct device ID as far as USB is concerned
	if(!HostSendRequest(dev, NULL, (USBStandardRequest) {
		.requestType = 0x00,
		.request = 5, // SET_ADDRESS
		.value = tmp,
		.index = 0,
		.length = 0
	})) {
		dev->id = tmp;
		return 0;
	}
	dev->id = tmp;
	
	USBDeviceDescriptor ddesc;
	if(!HostSendRequest(dev, &ddesc, (USBStandardRequest) {
		.requestType = 0x80,
		.request = 6, // GET_DESCRIPTOR
		.value = 1 << 8, // device
		.index = 0,
		.length = 8
	})) return 0;
	
	dev->maxPacketSize = ddesc.maxPacketSize;
	
	if(ddesc.clazz == 0 && ddesc.subclazz == 0 && ddesc.protocol == 0) {
		uint8_t buf[64];
		if(HostSendRequest(dev, buf, (USBStandardRequest) {
			.requestType = 0x80,
			.request = 6, // GET_DESCRIPTOR
			.value = 2 << 8, // configuration
			.index = 0,
			.length = sizeof(buf)
		})) {
			USBConfigurationDescriptor *conf = (void*) buf;
			
			if(HostSendRequest(dev, NULL, (USBStandardRequest) {
				.requestType = 0x00,
				.request = 9,
				.value = conf->id,
				.index = 0,
				.length = 0
			})) {
				if(conf->interfaceCount == 1) {
					USBInterfaceDescriptor *interf = (USBInterfaceDescriptor*) ((uintptr_t) conf + conf->sz);
					
					if(HostSendRequest(dev, NULL, (USBStandardRequest) {
						.requestType = 0x01,
						.request = 11,
						.value = interf->alternate,
						.index = interf->interface,
						.length = 0
					})) {
						if(interf->clazz == 0x08 && interf->subclazz == 0x06 && interf->protocol == 0x50) {
							USBEndpointDescriptor *endp0 = (USBEndpointDescriptor*) ((uintptr_t) interf + interf->sz);
							USBEndpointDescriptor *endp1 = (USBEndpointDescriptor*) ((uintptr_t) endp0 + endp0->sz);

							dev->bbb.maxPacketSize = endp0->maxPacketSize;
							dev->bbb.toggleIn = dev->bbb.toggleOut = 0;
							
							if(endp0->endpoint & 0x80) {
								dev->bbb.endpIn = endp0->endpoint;
								dev->bbb.endpOut = endp1->endpoint;
							} else {
								dev->bbb.endpIn = endp1->endpoint;
								dev->bbb.endpOut = endp0->endpoint;
							}
							
							uint8_t lun;
							if(!HostSendRequest(dev, &lun, (USBStandardRequest) {
								.requestType = 0xA1,
								.request = 0xFE, // get max lun
								.value = 0,
								.index = 0,
								.length = sizeof(lun)
							})) lun = 0;
							
							// (Almost) initialization procedure laid out by Doug Brown
							
							int attempts;
							
							attempts = 100;
							
							while(attempts-- > 0) {
								
								int success = 1;
								
								if(!BBBTestUnitReady(dev)) {
									success = 0;
								}
								
								if(!BBBInquiry(dev)) {
									success = 0;
								}
								
								if(success) {
									break;
								}
								
							}
							
							if(attempts > 0) {
								
								attempts = 100;
								
								while(attempts-- > 0) {
									if(BBBReadCapacity10(dev)) {
										break;
									}
								}
								
								if(attempts > 0) {
									
									//BBBModeSense(dev);
									
									// One flash drive doesn't work at all without this Read(10)
									uint8_t lolibuf[dev->bbb.sectorSize];
									BBBRead10(dev, 0, lolibuf, 1);
									
									for(uint32_t try = 0; try <= 0xFFFF; try++) {
										if((dev->canal = sys_canal_create('STRG', 'UD\0\0' | try)) < (uint32_t) -256) {
											break;
										}
									}
									
									if(dev->canal >= (uint32_t) -256) {
										return 0;
									}
									
									dev->respond = BBBDeviceResponse;
									
								}
								
							}
						}
					}
				}
			}
		}
	}
	
	return 1;
}
