#include"kmeans.h"

#include"../sys.h"

#define RED(c) (c >> 11)
#define GRN(c) ((c >> 5) & 63)
#define BLU(c) (c & 31)

#define CLUSTERS 256
struct {
	MiniCol averages[CLUSTERS];
} Cluster;

typedef struct {
	uint8_t cluster;
	MiniCol color;
} __attribute__((packed)) Datapoint;

struct {
	size_t sz;
	Datapoint *buf;
} Dataset = {};

__attribute__((optimize("Ofast"))) static uint32_t coldist(MiniCol a, MiniCol b) {
	uint16_t dr = RED(a) - RED(b);
	uint16_t dg = GRN(a) - GRN(b);
	uint16_t db = BLU(a) - BLU(b);
	return (uint16_t) (dr * dr) + (uint16_t) (dg * dg) + (uint16_t) (db * db);
}

void q_reavg() {
	if(Dataset.sz < CLUSTERS) {
		for(size_t i = 0; i < Dataset.sz; i++) {
			MiniCol c = Dataset.buf[i].color;
			Cluster.averages[Dataset.buf[i].cluster = i] = c;
		}
	} else {
		uint16_t counts[CLUSTERS] = {};
		uint16_t avgRed[CLUSTERS] = {};
		uint32_t avgGreen[CLUSTERS] = {};
		uint16_t avgBlue[CLUSTERS] = {};
		for(size_t i = 0; i < Dataset.sz; i++) {
			counts[Dataset.buf[i].cluster]++;
			avgRed[Dataset.buf[i].cluster] += RED(Dataset.buf[i].color);
			avgGreen[Dataset.buf[i].cluster] += GRN(Dataset.buf[i].color);
			avgBlue[Dataset.buf[i].cluster] += BLU(Dataset.buf[i].color);
		}
		
		for(int i = 0; i < CLUSTERS; i++) {
			if(counts[i] == 0) continue;
			
			avgRed[i] /= counts[i];
			avgGreen[i] /= counts[i];
			avgBlue[i] /= counts[i];
			
			Cluster.averages[i] = (avgRed[i] << 11) | (avgGreen[i] << 5) | avgBlue[i];
		}
	}
	
	for(size_t i = 0; i < Dataset.sz; i++) {
		Dataset.buf[i].cluster = q_p2c(Dataset.buf[i].color);
	}
}

__attribute__((optimize("Ofast"))) int q_p2c(MiniCol col) {
	int clu = 0;
	uint32_t dist = coldist(col, Cluster.averages[clu]);
	
	for(int i = 1; i < CLUSTERS; i++) {
		uint32_t d = coldist(col, Cluster.averages[i]);
		if(dist > d) {
			dist = d;
			clu = i;
		}
	}
	
	return clu;
}

void q_reassign() {
	asm("out %%al, %%dx" : : "a"(0), "d"(0x3C8));
	size_t bound = Dataset.sz < CLUSTERS ? Dataset.sz : CLUSTERS;
	for(size_t i = 0; i < bound; i++) {
		MiniCol c = Cluster.averages[i];
		asm("out %%al, %%dx" : : "a"((c >> 10) & ~1), "d"(0x3C9));
		asm("out %%al, %%dx" : : "a"((c >> 5) & 63), "d"(0x3C9));
		asm("out %%al, %%dx" : : "a"((c & 31) << 1), "d"(0x3C9));
	}
}

void q_addp(MiniCol col) {
	if(Dataset.sz == 4096 / sizeof(Datapoint)) {
		return;
	}
	
	Dataset.buf[Dataset.sz++] = (Datapoint) {.cluster = q_p2c(col), .color = col};
}

void q_init() {
	Dataset.buf = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 4096);
}