#pragma once

#include"../vid/api.h"
#include"../sys.h"

extern void *VidLink;

static inline void immdraw_font_draw(uint16_t x, uint16_t y, uint16_t w, size_t txtlen, const char *txt, uint8_t align) {
	LumaVidCommand *cmd = VidLink;
	
	cmd->op = LUMA_VID_COMMAND_SET_COLOR_FLAT;
	cmd->colorFlat.r = 0x00;
	cmd->colorFlat.g = 0x00;
	cmd->colorFlat.b = 0x00;
	cmd = (LumaVidCommand*) ((uintptr_t) cmd + sizeof(cmd->colorFlat));
	
	cmd->op = LUMA_VID_COMMAND_TEXT;
	cmd->text.x = x;
	cmd->text.y = y;
	cmd->text.wall = x + w;
	cmd->text.align = align;
	cmd->text.len = txtlen;
	std_copy(cmd->text.data, txt, txtlen);
	cmd = (LumaVidCommand*) ((uintptr_t) cmd + sizeof(cmd->text) + txtlen);
	
	cmd->op = LUMA_VID_COMMAND_END;
	
	memory_barrier();
	sys_signal_send(VidLink, -1);
	sys_signal_wait(VidLink, NULL);
}

static inline void immdraw_fill_rect(uint16_t x, uint16_t y, uint16_t w, uint16_t h, uint8_t r, uint8_t g, uint8_t b) {
	LumaVidCommand *cmd = VidLink;
	
	cmd->op = LUMA_VID_COMMAND_SET_COLOR_FLAT;
	cmd->colorFlat.r = r;
	cmd->colorFlat.g = g;
	cmd->colorFlat.b = b;
	cmd = (LumaVidCommand*) ((uintptr_t) cmd + sizeof(cmd->colorFlat));
	
	cmd->op = LUMA_VID_COMMAND_RECT;
	cmd->rect.x1 = x;
	cmd->rect.x2 = x + w - 1;
	cmd->rect.y1 = y;
	cmd->rect.y2 = y + h - 1;
	cmd = (LumaVidCommand*) ((uintptr_t) cmd + sizeof(cmd->rect));
	
	cmd->op = LUMA_VID_COMMAND_END;
	
	memory_barrier();
	sys_signal_send(VidLink, -1);
	sys_signal_wait(VidLink, NULL);
}

static inline void immdraw_text_extent(size_t txtlen, const uint8_t *txt, size_t wall, uint16_t *w, uint16_t *h) {
	LumaVidCommand *cmd = VidLink;
	
	cmd->op = LUMA_VID_COMMAND_TEXT_EXTENT;
	cmd->textExtent.wall = wall;
	cmd->textExtent.len = txtlen;
	std_copy(cmd->textExtent.data, txt, txtlen);
	cmd = (LumaVidCommand*) ((uintptr_t) cmd + sizeof(cmd->textExtent) + txtlen);
	
	cmd->op = LUMA_VID_COMMAND_END;
	
	memory_barrier();
	sys_signal_send(VidLink, -1);
	sys_signal_wait(VidLink, NULL);
	
	*w = cmd->textExtent.w;
	*h = cmd->textExtent.h;
}
