#include"perform.h"

#include"../fs/fs.h"
#include"wm.h"
#include"../sys.h"
#include"logserver.h"
#include"progdb.h"
#include"eebie/reader.h"
#include"tiny.h"
#include"io.h"
#include"fileswindow.h"
#include"k3menu.h"

__attribute__((optimize("Ofast"))) static void b2hex(uint8_t b, char *ret) {
	uint8_t top = b >> 4;
	uint8_t bottom = b & 0x0F;
	
	ret[0] = top < 0xA ? (top + '0' - 0) : (top + 'A' - 0xA);
	ret[1] = bottom < 0xA ? (bottom + '0' - 0) : (bottom + 'A' - 0xA);
}

static void show_help_window() {
	if(!is_free_space()) return;
	
	/*static const char txt[] = "Luma's interface is designed around the keyboard. In fact, you are unlikely to find any use for a mouse except in niche cases.\n\nTurn to the command bar below by pressing the System (Sys) key. Start off by exploring Luma's software by entering \"list\" or your filesystem with \"files\".\n\nSys+H and Sys+V split the desktop to fit more windows. Use Sys+Escape to close windows or remove free space. Sys+P moves windows around.";
	struct GLabel *l = glabel_new(0, 0, 0, 0, g_dynstr16_new(sizeof(txt) - 1, sizeof(txt) - 1, txt), 1);
	
	static const char name[] = "Welcome to Luma";
	struct GWindow *w = gwindow_new(0, 0, 0, 0, g_dynstr16_new(sizeof(name) - 1, sizeof(name) - 1, name), (struct GObj*) l, 0);
	
	add_window(w);*/
	
	struct GWindow *gw = add_window("Welcome to Luma");
	
	struct k3MLabel *lbl = k3MLabel(NULL, 14, "Luma's interface is designed around the keyboard. In fact, you are unlikely to find any use for a mouse except in niche cases.\n\nTurn to the command bar below by pressing the System (Sys) key. Start off by exploring Luma's software by entering \"list\" or your filesystem with \"files\".\n\nSys+H and Sys+V split the desktop to fit more windows. Use Sys+Escape to close windows or remove free space. Sys+P moves windows around.");
	k3MenuSetBounds(lbl, 0, 20, 320, 204);
	k3MAddChild(gw->panel, lbl);
	
	dirty_obj(gw->whole);
}

static void show_log_window() {
	if(!is_free_space()) return;
	
	/*struct GLabel *l = glabel_new(0, 0, 0, 0, log_get(), 0);
	
	static const char name[] = "System Logs";
	struct GWindow *w = gwindow_new(0, 0, 0, 0, g_dynstr16_new(sizeof(name) - 1, sizeof(name) - 1, name), (struct GObj*) l, 0);
	
	add_window(w);*/
}

static void e9(const char *buf, size_t len) {
	while(len--) {
		asm volatile("out %%al, $0xE9" :: "a"(*buf) :);
		buf++;
	}
}

static void create_process_from_file(const uint8_t *exefile, size_t exesize) {
	size_t childHandle = sys_spawn();
	
	EBMLReader *rdr = tiny_malloc(sizeof(*rdr));
	ebml_reader_init(rdr);
	
	struct Section {
		uint8_t *data;
		size_t dataSz;
		
		uint8_t *their;
		
		uint8_t *relocs;
		size_t relocsSz;
	};
	
	struct Ctx {
		uint8_t *buf;
		size_t size;
		
		struct Section section[4];
		int sIdx;
		
		int currentSymbolIsEntry;
		
		int entrySection;
		size_t entryOffset;
	} ctx = {};
	
	ctx.entrySection = -1; // i.e. no entry symbol
	
	rdr->ud = &ctx;
	
	EBMLElementType eventer(EBMLReader *rdr, uint64_t id, uint64_t length) {
		struct Ctx *ctx = rdr->ud;
		
		ctx->buf = NULL;
		ctx->size = 0;
		
		if(id == 0x10000000 || id == 0x6200 || id == 0xC0) {
			return EBML_TREE;
		}
		
		return EBML_BINARY;
	}
	void evchu(EBMLReader *rdr, const uint8_t *data, size_t length) {
		struct Ctx *ctx = rdr->ud;
		
		ctx->buf = tiny_realloc(ctx->buf, ctx->size + length);
		std_copy(ctx->buf + ctx->size, data, length);
		ctx->size += length;
	}
	void evexit(EBMLReader *rdr) {
		struct Ctx *ctx = rdr->ud;
		
		size_t exitingId = rdr->idStack[rdr->currentDepth - 1];
		
		if(exitingId == 0x6201) {
			ctx->section[ctx->sIdx].data = ctx->buf;
			ctx->section[ctx->sIdx].dataSz = ctx->size;
			
			ctx->buf = NULL;
		} else if(exitingId == 0x6202) {
			ctx->section[ctx->sIdx].relocs = ctx->buf;
			ctx->section[ctx->sIdx].relocsSz = ctx->size;
			
			ctx->buf = NULL;
		} else if(exitingId == 0x6200) {
			uint8_t *my = SYS_MAP_VIRT_ANY;
			uint8_t *their = SYS_MAP_VIRT_ANY;
			
			sys_abuse_map(childHandle, &my, &their, SYS_MAP_PHYS_ANY, ctx->section[ctx->sIdx].dataSz);
			
			std_copy(my, ctx->section[ctx->sIdx].data, ctx->section[ctx->sIdx].dataSz);
			
			tiny_free(ctx->section[ctx->sIdx].data);
			
			ctx->section[ctx->sIdx].data = my;
			ctx->section[ctx->sIdx].their = their;
			
			ctx->sIdx++;
		}
		
		if(exitingId == 0xC0) {
			ctx->currentSymbolIsEntry = 0;
		} else if(exitingId == 0xC1) {
			if(ctx->size == 12) {
				if(ctx->buf[0] == 'P' && ctx->buf[1] == 'r' && ctx->buf[2] == 'o' && ctx->buf[3] == 'g' && ctx->buf[4] == 'r' && ctx->buf[5] == 'a' && ctx->buf[6] == 'm' && ctx->buf[7] == 'E' && ctx->buf[8] == 'n' && ctx->buf[9] == 't'
					&& ctx->buf[10] == 'r' && ctx->buf[11] == 'y') {
					
					ctx->currentSymbolIsEntry = 1;
				}
			}
		} else if(exitingId == 0xC2) {
			if(ctx->currentSymbolIsEntry) {
				ctx->entryOffset = *(uint32_t*) &ctx->buf[0];
				ctx->entrySection = *(uint8_t*) &ctx->buf[4];
			}
		}
		
		if(ctx->buf) {
			tiny_free(ctx->buf);
		}
	}
	
	rdr->eventEnterElement = eventer;
	rdr->eventDataChunk = evchu;
	rdr->eventExitElement = evexit;
	
	uint8_t *exefilestart = exefile;
	
	while(1) {
		int readCount = ebml_reader_feed(rdr, exefile, exesize);
		
		if(readCount <= 0) {
			break;
		}
		
		exefile += readCount;
		exesize -= readCount;
	}
	
	// Perform relocations
	for(int s = 0; s < ctx.sIdx; s++) {
		for(int r = 0; r < ctx.section[s].relocsSz / 5; r++) {
			size_t offset = *(uint32_t*) &ctx.section[s].relocs[r * 5 + 0];
			int refSec = *(uint8_t*) &ctx.section[s].relocs[r * 5 + 4];
			
			*(uint32_t*) &ctx.section[s].data[offset] += (uint32_t) ctx.section[refSec].their;
		}
		
		tiny_free(ctx.section[s].relocs);
		ctx.section[s].relocs = NULL;
		
		sys_vpm_unmap(ctx.section[s].data, ctx.section[s].dataSz); // I'm scared this will break
	}
	
	sys_abuse_state(childHandle, SYS_STATE_X86_EIP, ctx.section[ctx.entrySection].their + ctx.entryOffset);
	
	sys_release(childHandle);
	
	tiny_free(rdr);
}

static void attempt_program_launch(const uint8_t *str, size_t len) {
	size_t p;
	
	for(p = 0; p < ProgramDB.count; p++) {
		Str16 *name = ProgramDB.names[p];
		
		uint8_t *displayname = name->data;
		uint8_t *codename = std_bytefind(name->len, name->data, ';') + 1;
		
		size_t testlen = name->len - (codename - displayname);
		if(testlen > len) {
			testlen = len;
		}
		
		if(std_bytecomp(testlen, codename, str) == 0) {
			break;
		}
	}
	
	if(p == ProgramDB.count) {
		return;
	}
	
	size_t exesize;
	uint8_t *exefile;
	{
		Str16 *name = ProgramDB.names[p];
		
		uint8_t *displayname = name->data;
		uint8_t *term = std_bytefind(name->len, name->data, ';');
		
		Str16 *directory = __builtin_alloca(sizeof(*directory) + (term - displayname) + sizeof(*directory) + 7 + sizeof(*directory));
		std_copy(directory->data, displayname, term - displayname);
		directory->len = term - displayname;
		
		Str16 *fn = WALK_STR16(directory);
		fn->len = 7;
		fn->data[0] = 'P';
		fn->data[1] = 'r';
		fn->data[2] = 'o';
		fn->data[3] = 'g';
		fn->data[4] = 'r';
		fn->data[5] = 'a';
		fn->data[6] = 'm';
		
		// Terminator.
		WALK_STR16(fn)->len = 0;
		
		exefile = get_file_from_path(directory, &exesize);
	}
	
	create_process_from_file(exefile, exesize);
	
	tiny_free(exefile);
	
	/*size_t handle = sys_spawn();
	
	uint8_t *code = sys_abuse_map(handle, SYS_MAP_VIRT_ANY, 0x40000000, SYS_MAP_PHYS_ANY, 4096);
	code[0] = 0x66;
	code[1] = 0x87;
	code[2] = 0xDB;
	code[3] = 0xEB;
	code[4] = 0xFB;
	
	sys_abuse_state(handle, SYS_STATE_X86_EIP, 0x40000000);
	
	sys_release(handle);*/
}

void perform(DynStr16 *cmd) {
	if(cmd->len == 4 && cmd->data[0] == 'h' && cmd->data[1] == 'e' && cmd->data[2] == 'l' && cmd->data[3] == 'p') {
		show_help_window();
	} else if(cmd->len == 5 && cmd->data[0] == 'f' && cmd->data[1] == 'i' && cmd->data[2] == 'l' && cmd->data[3] == 'e' && cmd->data[4] == 's') {
		FilesWindowCreate();
	} else if(cmd->len == 4 && cmd->data[0] == 'l' && cmd->data[1] == 'o' && cmd->data[2] == 'g' && cmd->data[3] == 's') {
		show_log_window();
	} else if(cmd->len > 2 && cmd->data[0] == 'g' && cmd->data[1] == 'o' && cmd->data[2] == ' ') {
		attempt_program_launch(cmd->data + 3, cmd->len - 3);
	} /*else if(windowTreeFocus && windowTreeFocus->window) {
		uint8_t *space = (uint8_t*) std_bytefind(cmd->len, (char*) cmd->data, ' ');
		
		size_t end = space ? space - cmd->data : cmd->len;
		for(size_t i = 0; i < windowTreeFocus->window->cmds.count; i++) {
			if(windowTreeFocus->window->cmds.array[i].name->len != end) {
				continue;
			}
			
			if(std_bytecomp(end, cmd->data, windowTreeFocus->window->cmds.array[i].name->data) != 0) {
				continue;
			}
			
			windowTreeFocus->window->cmds.array[i].run(windowTreeFocus->window->cmds.array[i].ud, cmd);
			
			break;
		}
	}*/ else {
		// Failed.
	}
}
