cpu 386
bits 32

TIMER_FREQUENCY equ 250;Hz

; TODO: perform IO waits

section .text

extern scheduler_switch
extern scheduler_CurrentProcess
pit_handler:
extern scheduler_Tick
	inc dword [scheduler_Tick]
	push eax
	mov al, 0x20
	out 0x20, al
	pop eax
	
ctx_switch:
	push eax
	
	mov eax, [scheduler_CurrentProcess]
	test eax, eax
	jnz .a
	add esp, 16
	jmp .b
.a:
	pop dword [eax + 2] ;eax
	pop dword [eax + 34] ;eip
	pop dword [eax + 42] ;cs
	pop dword [eax + 38] ;eflags
	mov [eax + 6], ebx
	mov [eax + 10], ecx
	mov [eax + 14], edx
	mov [eax + 18], esi
	mov [eax + 22], edi
	mov [eax + 26], ebp
	cmp dword [eax + 42], 8 ;CS 8 implies ring 0
	jne .isR3
	mov [eax + 30], esp
	jmp .b
.isR3:
	pop dword [eax + 30] ;esp
	add esp, 4 ;ss
.b:
	call scheduler_switch
	jmp jump_to_process

global jump_to_process
jump_to_process:
	mov eax, [scheduler_CurrentProcess]
	test eax, eax
	jnz .c
	; No next process, goto hlt loop.
	push dword 0x202
	push dword 8
	push haltLoop
	iretd
.c:
	mov ebx, [eax + 46]
	mov cr3, ebx
	mov ebx, [eax + 6]
	mov ecx, [eax + 10]
	mov edx, [eax + 14]
	mov esi, [eax + 18]
	mov edi, [eax + 22]
	mov ebp, [eax + 26]
	cmp dword [eax + 42], 8
	je .isR0
	push dword 32 | 3
	push dword [eax + 30]
	jmp .b
.isR0:
	mov esp, [eax + 30]
.b:
	push dword [eax + 38] ;eflags
	push dword [eax + 42] ;cs
	push dword [eax + 34] ;eip
	mov eax, [eax + 2]
	iretd

set_bit:
	push edx
	push ebx
	push ecx
	push eax
	
	shl ebx, 1
	
	mov dx, 0x3C4
	mov al, 2
	out dx, al
	
	mov ecx, ebx
	and ecx, 3
	mov eax, 1
	shl eax, cl
	
	mov dx, 0x3C5
	out dx, al
	
	mov ecx, ebx
	sar ecx, 2
	
	pop eax
	
	mov [0xA0000 + ecx], byte al
	
	pop ecx
	pop ebx
	pop edx
	
	ret

write_dword:
	push eax
	push ecx
	
	mov ecx, 32
	
.lup:
	
	rol eax, 1
	
	push eax
	and eax, 1
	shl eax, 4
	add eax, 20
	call set_bit
	pop eax
.nz:
	inc ebx
	
	dec ecx
	jnz .lup
	
	pop ecx
	pop eax
	
	ret

clear_screen:
	push eax
	push ecx
	push edx
	push edi
	
	mov dx, 0x3C4
	mov al, 2
	out dx, al
	
	inc dx
	mov al, 15
	out dx, al
	
	mov edi, 0xA0000
	mov ecx, 4800
	mov eax, 0
	rep stosd
	
	pop edi
	pop edx
	pop ecx
	pop eax
	
	ret

pgf_handler:
	call clear_screen
	
	pop eax ;error code
	mov ebx, 0
	call write_dword
	
	inc ebx
	
	mov eax, cr2 ;accessed address
	call write_dword
	
	inc ebx
	
	pop eax ;instruction cause of exception
	call write_dword
	
	add ebx, 100
	
	mov eax, [eax] ;instruction data
	call write_dword
.LOUP:
	hlt
	jmp .LOUP

dbf_handler:
	add esp, 4
	mov [0xA0000], dword "D_F_"
.LOUP:
	hlt
	jmp .LOUP

gpf_handler:
	pop eax
	movzx ebx, al
	and bl, 0xF
	mov bl, [.tbl + ebx]
	mov [0xB800E], bl
	movzx ebx, al
	shr bl, 4
	mov bl, [.tbl + ebx]
	mov [0xB800C], bl
	shr eax, 8
	movzx ebx, al
	and bl, 0xF
	mov bl, [.tbl + ebx]
	mov [0xB800A], bl
	movzx ebx, al
	shr bl, 4
	mov bl, [.tbl + ebx]
	mov [0xB8008], bl
	shr eax, 8
	movzx ebx, al
	and bl, 0xF
	mov bl, [.tbl + ebx]
	mov [0xB8006], bl
	movzx ebx, al
	shr bl, 4
	mov bl, [.tbl + ebx]
	mov [0xB8004], bl
	shr eax, 8
	movzx ebx, al
	and bl, 0xF
	mov bl, [.tbl + ebx]
	mov [0xB8002], bl
	movzx ebx, al
	shr bl, 4
	mov bl, [.tbl + ebx]
	mov [0xB8000], bl
	shr eax, 8
.LOUP:
	hlt
	jmp .LOUP
.tbl: db "0123456789ABCDEF"

spur_handler:
	iret

syscall_int_entry:
	cmp eax, (.tblend - .tbl) / 4
	jae .invalid
	jmp [.tbl + eax * 4]
.invalid:
	mov eax, 0
	iret
.tbl:
	dd .sys_vpm_map, .sys_canal_create, .sys_link_create, .sys_canal_accept, .sys_signal_send, .sys_signal_wait, .sys_pci_claim, .sys_pci_read, .sys_vpm_unmap, .sys_vpm_get_phys, .sys_sleep, .sys_pci_write_1, .sys_pci_write_2, .sys_pci_write_4, .sys_irq_claim, .sys_wait_for_irq, .sys_create_thread, .sys_spawn, .sys_abuse_map, .sys_release, .sys_abuse_state, .sys_exit, .sys_event_wait
.tblend:
.sys_vpm_map:
	push ecx
	push edx
	push esi
	push edi
	push 1 ;USER mapping
	mov eax, cr3
	push eax
extern vpm_map
	call vpm_map
	add esp, 16
	pop edx
	pop ecx
	iret
.sys_canal_create:
	push ecx
	push edx
	mov eax, [scheduler_CurrentProcess]
	movzx eax, word [eax] ;process ID
	push eax
	push esi
	push edi
extern canal_create
	call canal_create
	add esp, 12
	pop edx
	pop ecx
	iret
.sys_link_create:
	push ecx
	push edx
	mov eax, [scheduler_CurrentProcess]
	movzx eax, word [eax]
	push eax
	push esi
	push edi
extern link_create
	call link_create
	add esp, 12
	pop edx
	pop ecx
	iret
.sys_canal_accept:
	push ecx
	push edx
	push edi
	mov eax, [scheduler_CurrentProcess]
	movzx eax, word [eax] ;id
	push eax
extern canal_accept
	call canal_accept
	add esp, 8
	pop edx
	pop ecx
	iret
.sys_signal_send:
	push ecx
	push edx
	push esi
	push edi
	mov eax, [scheduler_CurrentProcess]
	movzx eax, word [eax] ;id
	push eax
extern link_send_signal_from
	call link_send_signal_from
	pop eax
	add esp, 8
	pop edx
	pop ecx
	iret
.sys_signal_wait:
	push eax
	mov eax, [scheduler_CurrentProcess]
	mov [eax + 50], byte 1
	mov [eax + 51], edi
	pop eax
	jmp ctx_switch
.sys_pci_claim:
	push ecx
	push edx
	push edi
extern pci_claim
	call pci_claim
	pop edx
	pop edx
	pop ecx
	iret
.sys_pci_read:
	push ecx
	push edx
	push esi
	push edi
extern pci_read
	call pci_read
	add esp, 8
	pop edx
	pop ecx
	iret
.sys_vpm_unmap:
	push eax
	push ecx
	push edx
	push esi
	push edi
	mov eax, cr3
	push eax
extern vpm_unmap
	call vpm_unmap
	add esp, 12
	pop edx
	pop ecx
	pop eax
	iret
.sys_vpm_get_phys:
	push ecx
	push edx
	push edi
	mov eax, cr3
	push eax
extern vpm_get_phys
	call vpm_get_phys
	add esp, 8
	pop edx
	pop ecx
	iret
.sys_sleep:
	push edx
	push ebx
	push eax
	xor edx, edx
	mov eax, edi
	mov ebx, 1000 / TIMER_FREQUENCY
	div ebx
	add eax, [scheduler_Tick]
	mov ebx, [scheduler_CurrentProcess]
	mov byte [ebx + 50], 2
	mov [ebx + 51], eax
	pop eax
	pop ebx
	pop edx
	jmp ctx_switch
.sys_pci_write_1:
	push eax
	push ecx
	push edx
	push esi
	push edi
extern pci_write_1
	call pci_write_1
	add esp, 8
	pop edx
	pop ecx
	pop eax
	iret
.sys_pci_write_2:
	push eax
	push ecx
	push edx
	push esi
	push edi
extern pci_write_2
	call pci_write_2
	add esp, 8
	pop edx
	pop ecx
	pop eax
	iret
.sys_pci_write_4:
	push eax
	push ecx
	push edx
	push esi
	push edi
extern pci_write_4
	call pci_write_4
	add esp, 8
	pop edx
	pop ecx
	pop eax
	iret
.sys_irq_claim:
	push ecx
	push edx
	call irq_claim
	pop edx
	pop ecx
	iret
.sys_wait_for_irq:
	push ecx
	mov ecx, edi
	; TODO: check privilege
	call toggle_irq_mask
	mov ecx, [scheduler_CurrentProcess]
	mov [ecx + 50], byte 3 ; Wait type IRQ
	pop ecx
	jmp ctx_switch
.sys_create_thread:
	push eax
	push ecx
	push edx
extern scheduler_spawn
	call scheduler_spawn
	mov ecx, cr3
	mov [eax + 46], ecx ;cr3
	mov [eax + 34], edi ;eip
	pop edx
	pop ecx
	pop eax
	mov eax, 1
	iret
.sys_spawn:
	push ecx
	push edx
	call scheduler_spawn
	test eax, eax
	jnz .sys_spawn_spawned
	pop edx
	pop ecx
	iret
.sys_spawn_spawned:
	push eax
	mov eax, [scheduler_CurrentProcess]
	movzx eax, word [eax] ;proc ID
	push eax
extern scheduler_enslave
	call scheduler_enslave
	add esp, 8
	pop edx
	pop ecx
	iret
.sys_abuse_map:
	push ecx
	push edx
		push edi
		call scheduler_find
		add esp, 4
	pop edx
	pop ecx
	
	test eax, eax
	jz .notfound
	
	push ecx
	push edx
		push ebx
		push ecx
		push edx
		push esi
		push dword [eax + 46] ;cr3
		mov eax, cr3
		push eax
extern vpm_double_map
		call vpm_double_map
		add esp, 24
	pop edx
	pop ecx
	
	iret
.notfound:
	mov eax, -1
	iret
.sys_release:
	push ecx
	push edx
	push edi
extern scheduler_unslave
	call scheduler_unslave
	pop edi
	pop edx
	pop ecx
	iret
.sys_abuse_state:
	push ecx
	push edx
		push edi
		call scheduler_find
		add esp, 4
	pop edx
	pop ecx
	
	test eax, eax
	jz .notfound2
	
	push ecx
	push edx
		push edx
		push esi
		push eax
extern process_abuse_state
		call process_abuse_state
		add esp, 12
	pop edx
	pop ecx
	
	iret
.notfound2:
	mov eax, -1
	iret
.sys_exit:
	iret
.sys_event_wait:
	push eax
	push ecx
	push edx
		push edi
		push dword [scheduler_CurrentProcess]
extern process_search_event
		call process_search_event
		add esp, 8
	pop edx
	pop ecx
	pop eax
	iret
	

global scheduler_start:
scheduler_start:
.pic_init:
	mov al, 0x10 | 0x01
	out 0x20, al
	out 0xA0, al
	
	mov al, 32
	out 0x21, al
	mov al, 40
	out 0xA1, al
	
	mov al, 4
	out 0x21, al
	mov al, 2
	out 0xA1, al
	
	mov al, 0x01
	out 0x21, al
	out 0xA1, al
	 
.pic_mask:
	mov al, ~1 ;PIT
	out 0x21, al
	mov al, ~0
	out 0xA1, al
	
.idt:
	; Page fault handler
	mov eax, pgf_handler
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 14 * 8], eax
	
	mov eax, pgf_handler
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + 14 * 8 + 4], eax
	
	; Double fault handler
	mov eax, dbf_handler
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 8 * 8], eax
	
	mov eax, dbf_handler
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + 8 * 8 + 4], eax
	
	; General protection fault handler
	mov eax, gpf_handler
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 13 * 8], eax
	
	mov eax, gpf_handler
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + 13 * 8 + 4], eax
	
	; Timer scheduler handler
	mov eax, pit_handler
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 32 * 8], eax
	
	mov eax, pit_handler
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + 32 * 8 + 4], eax
	
	; Spurious interrupt handler
	mov eax, spur_handler
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 39 * 8], eax
	
	mov eax, spur_handler
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + 39 * 8 + 4], eax
	
	; Software interrupt syscall entry point
	mov eax, syscall_int_entry
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + 0xEC * 8], eax
	
	mov eax, syscall_int_entry
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15) | (3 << 13)
	mov [idt + 0xEC * 8 + 4], eax
	
	push idt
	push (idt.end - idt - 1) << 16
	lidt [esp + 2]
	add esp, 8
	
.pit:
	mov al, (2 << 1) | (2 << 4)
	out 0x43, al
	
	mov al, (1193182 / TIMER_FREQUENCY) % 256
	out 0x40, al
	mov al, (1193182 / TIMER_FREQUENCY) / 256
	out 0x40, al
	
	sti
	
haltLoop:
	hlt
	jmp haltLoop

; ecx = id
toggle_irq_mask:
	push ecx
	push ebx
	push edx
	push eax
	mov dx, 0x21
	cmp cl, 8
	jl .lower
	sub cl, 8
	mov dx, 0xA1
.lower:
	mov bl, 1
	shl bl, cl
	in al, dx
	xor al, bl
	out dx, al
	pop eax
	pop edx
	pop ebx
	pop ecx
	ret

irq_claim:
	test edi, edi
	jnz .notTimer
	xor eax, eax
	ret
.notTimer:
	mov eax, [scheduler_CurrentProcess]
	mov bx, [eax] ;id
	lea eax, [edi * 8]
	lea eax, [eax * 2 + edi] ;17x
	lea eax, [eax + irqhandlers - 17]
	mov [eax + 2], esi ;mov edi, handler
	mov [eax + 7], bx ;mov esi, process id
	mov ecx, edi
	mov [eax + 12], cl
	
	push eax
	and eax, 0x0000FFFF
	or eax, 0x00080000
	mov [idt + edi * 8 + 256], eax
	
	pop eax
	and eax, 0xFFFF0000
	or eax, (14 << 8) | (1 << 15)
	mov [idt + edi * 8 + 256 + 4], eax
	
	mov eax, 1
	ret

irqhandlers:
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
	jmp strict word irqhandler
	pushad
	mov edi, strict dword 0
	mov esi, strict dword 0
	mov cl, strict byte 0
irqhandler:
	call toggle_irq_mask
	mov al, 0x20
	out 0x20, al
	push esi
extern scheduler_find
	call scheduler_find
	add esp, 4
	mov [eax + 55], edi
	popad
	iret

idt:
	times 256 dq 0
.end:
