#include"vpm.h"

#include"ppm.h"
#include"mem.h"

extern void *_kernel_end;

void set_temporary_mapping(uint32_t addr) {
	if((addr & 0xFFF) != 0) asm volatile("xchgw %bx, %bx");
	
	*(volatile uint32_t*) 0xFFFFFFF8 = addr | 3;
	asm volatile("invlpg 0xFFFFE000" : : : "memory");
}

static uint32_t nextAllocs[2] = {[VPM_KERNEL] = 0, [VPM_USER] = 1 * 1024 * 1024 * 1024};

void vpm_init(uint32_t kernelEnd) {
	nextAllocs[VPM_KERNEL] = (kernelEnd + 4095) & ~4095;
}

/* TODO: Make better. */
uint32_t vpm_find_free(uint32_t addrspace, AllocationOwner owner, size_t lenBytes) {
	size_t len = (lenBytes + 4095) / 4096;
	
	uint32_t ret = nextAllocs[owner];
	for(size_t i = 0; i < len;) {
		uint32_t pdeid = (ret + i * 4096) >> 22;
		uint32_t pteid = ((ret + i * 4096) >> 12) & 1023;
		
		set_temporary_mapping(addrspace);
		
		volatile uint32_t *pde = &((volatile uint32_t*) 0xFFFFE000)[pdeid];
		if((*pde & 1) == 0) {
			i += 4096;
		} else {
			set_temporary_mapping(*pde & ~0xFFF);
			
			if((((volatile uint32_t*) 0xFFFFE000)[pteid] & 1) == 0) {
				i++;
			} else {
				ret += (i + 1) * 4096;
				i = 0;
			}
		}
	}
	
	return ret;
}

/* If phys is -1, map to uncontiguous physical pages anywhere, else map contiguous span of physical memory. */
/* If phys is -2, automatically find contiguous span of physical memory */
/* If virt is -1, find free space, else overwrite. */
void *vpm_map(uint32_t addrspace, AllocationOwner owner, uint32_t virt, uint32_t phys, size_t len) {
	uint32_t thisspace;
	asm volatile("movl %%cr3, %0" : "=r" (thisspace) :);
	
	//if(len == 32768)asm("xchg %%bx, %%bx":::);
	
	if(virt == -1) {
		virt = vpm_find_free(addrspace, owner, len);
	}
	
	len = (len + 4095) / 4096;
	
	if(phys == -2) {
		phys = ppm_alloc(len);
		
		if(!phys) {
			goto error;
		}
	}
	
	for(size_t i = 0; i < len; i++) {
		set_temporary_mapping(addrspace);
		
		uint32_t pdeid = (virt + i * 4096) >> 22;
		uint32_t pteid = ((virt + i * 4096) >> 12) & 1023;
		
		volatile uint32_t *pde = &((volatile uint32_t*) 0xFFFFE000)[pdeid];
		
		if((*pde & 1) == 0) {
			uint64_t physpag = ppm_alloc(1);
			
			if(physpag) {
				*pde = physpag | 3 | (owner == VPM_USER ? 4 : 0);
			} else {
				goto error;
			}
		}
		set_temporary_mapping(*pde & ~0xFFF);
		
		volatile uint32_t *pte = &((volatile uint32_t*) 0xFFFFE000)[pteid];
		if(phys == -1) {
			*pte = ppm_alloc(1);
			if(!*pte) goto error;
		} else {
			*pte = phys + i * 4096;
		}
		*pte |= 3 | (owner == VPM_USER ? 4 : 0);
		
		if(thisspace == addrspace) {
			//~ asm volatile("xchg %bx, %bx");
			asm volatile("invlpg %0" : : "m"(*(char*) (virt + i * 4096)) : "memory");
		}
	}
	
	return (void*) virt;

error:
	// TODO: Deallocate physical pages.
	
	return 0;
}

int vpm_double_map(uint32_t cr3a, uint32_t cr3b, uint32_t *va, uint32_t *vb, uintptr_t phys, size_t len) {
	uint32_t thisspace;
	asm volatile("movl %%cr3, %0" : "=r" (thisspace) :);
	
	if(*va == -1) {
		*va = vpm_find_free(cr3a, VPM_USER, len);
	}
	if(*vb == -1) {
		*vb = vpm_find_free(cr3b, VPM_USER, len);
	}
	
	len = (len + 4095) / 4096;
	
	if(phys == -2) {
		phys = ppm_alloc(len);
	}
	
	for(size_t i = 0; i < len; i++) {
		uint32_t backend;
		
		if(phys == -1) {
			backend = ppm_alloc(1);
			if(!backend) goto error;
		} else {
			backend = phys + i * 4096;
		}
		
		{
			set_temporary_mapping(cr3a);
			
			uint32_t pdeid = (*va + i * 4096) >> 22;
			uint32_t pteid = ((*va + i * 4096) >> 12) & 1023;
			
			volatile uint32_t *pde = &((volatile uint32_t*) 0xFFFFE000)[pdeid];
			
			if((*pde & 1) == 0) {
				uintptr_t physpag = ppm_alloc(1);
				
				if(physpag) {
					*pde = physpag | 3 | 4;
				} else {
					goto error;
				}
			}
			
			set_temporary_mapping(*pde & ~0xFFF);
			
			volatile uint32_t *pte = &((volatile uint32_t*) 0xFFFFE000)[pteid];
			
			*pte = backend | 3 | 4;
			
			if(thisspace == cr3a) {
				asm volatile("invlpg %0" : : "m"(*(char*) (*va + i * 4096)) : "memory");
			}
		}
		
		{
			set_temporary_mapping(cr3b);
			
			uint32_t pdeid = (*vb + i * 4096) >> 22;
			uint32_t pteid = ((*vb + i * 4096) >> 12) & 1023;
			
			volatile uint32_t *pde = &((volatile uint32_t*) 0xFFFFE000)[pdeid];
			
			if((*pde & 1) == 0) {
				uintptr_t physpag = ppm_alloc(1);
				
				if(physpag) {
					*pde = physpag | 3 | 4;
				} else {
					goto error;
				}
			}
			
			set_temporary_mapping(*pde & ~0xFFF);
			
			volatile uint32_t *pte = &((volatile uint32_t*) 0xFFFFE000)[pteid];
			
			*pte = backend | 3 | 4;
			
			if(thisspace == cr3b) {
				asm volatile("invlpg %0" : : "m"(*(char*) (*vb + i * 4096)) : "memory");
			}
		}
	}
	
	return 1;

error:
	return 0;
}

uint32_t vpm_create_space() {
	uint32_t curPD;
	asm volatile("movl %%cr3, %0" : "=r"(curPD) :);
	set_temporary_mapping(curPD);
	
	uint32_t pde0000 = *(volatile uint32_t*) 0xFFFFE000;
	uint32_t pde1023 = *(volatile uint32_t*) 0xFFFFEFFC;
	
	uint32_t newPD = ppm_alloc(1);
	
	set_temporary_mapping(newPD);
	
	*(volatile uint32_t*) 0xFFFFE000 = pde0000;
	kmemw4((void*) 0xFFFFE004, 0, 1022);
	*(volatile uint32_t*) 0xFFFFEFFC = pde1023;
	
	return newPD;
}

void vpm_set_space(uint32_t cr3) {
	asm volatile("movl %0, %%cr3" : : "r"(cr3));
}

uint32_t vpm_get_phys(uint32_t cr3, uint32_t v) {
	uint32_t offset = v & 4095;
	
	uint32_t pdeid = v >> 22;
	uint32_t pteid = (v >> 12) & 1023;
	
	set_temporary_mapping(cr3);
	set_temporary_mapping(((uint32_t*) 0xFFFFE000)[pdeid] & ~0xFFF);
	
	return (((uint32_t*) 0xFFFFE000)[pteid] & ~0xFFF) + offset;
}

void vpm_unmap(uint32_t cr3, uint32_t virt, size_t len) {
	uint32_t thisspace;
	asm volatile("movl %%cr3, %0" : "=r"(thisspace) :);
	
	len = (len + 4095) / 4096;
	
	for(size_t i = 0; i < len; i++) {
		set_temporary_mapping(cr3);
		
		uint32_t pdeid = (virt + i * 4096) >> 22;
		uint32_t pteid = ((virt + i * 4096) >> 12) & 1023;
		
		volatile uint32_t *pde = &((volatile uint32_t*) 0xFFFFE000)[pdeid];
		
		if((*pde & 1) == 0) {
			continue;
		}
		
		set_temporary_mapping(*pde & ~0xFFF);
		
		((volatile uint32_t*) 0xFFFFE000)[pteid] = 0;
		
		if(thisspace == cr3) {
			asm volatile("invlpg %0" : : "m"(*(char*) (virt + i * 4096)) : "memory");
		}
	}
}
